from matplotlib.mlab import \
    csd, psd, window_hanning, detrend_mean, detrend_none
from numpy import pi, square, log10, conj
import numpy as np


class TransferFunction( object ):

    def __init__( self, constant, poles, zeroes ):

        self.constant	= constant
        self.poles	= poles
        self.zeroes	= zeroes


    @classmethod
    def fromstream( cls, fp ):

        constant = None
        poles = None
        zeroes = None

        readto = None

        for line in fp.readlines():

            if line.startswith('ZEROS'):

                zeroes = [0] * (int(line[len('ZEROS'):]))
                readto = zeroes
                pos = 0

            elif line.startswith('POLES'):

                poles = [0] * int(line[len('POLES'):])
                readto = poles
                pos = 0

            elif line.startswith('CONSTANT'):

                constant = float(line[len('CONSTANT'):])

            else:

                x, y = line.split()

                readto[pos] = float(x) + 1j*float(y)
                pos += 1

        return cls(constant = constant, poles = poles, zeroes = zeroes)


    def get_value( self, freq ):

        i_omega = 2j * pi * freq
        result = self.constant

        for zero in self.zeroes:

            result *= (i_omega - zero)

        for pole in self.poles:

            result /= (i_omega - pole)

        return result


def interpft( ys, newsamplerate ):
    """
    Returns interpolated points array of size newsamplerate.
    """

    return np.fft.irfft(np.fft.rfft(ys), newsamplerate) * newsamplerate / len(ys)


def abs_squared( z ):
    """
    Returns modulus squared of a complex value.
    """

    assert(np.iscomplexobj(z))

    return square(z.real) + square(z.imag)


def get_cov( data, sample_rate ):
    """
    Returns covariance matrix for signals data
    """

    size = len(data)
    data = list(map(lambda xs: detrend_none(np.array(xs)), data))

    params = {
        'NFFT'		: 64000,
        'window'	: window_hanning,
        'noverlap'	: 56000,
        'Fs'		: sample_rate,
        'detrend'	: detrend_none,
        'scale_by_freq'	: True,
        }

    # Create NxN array
    result = []

    for i in range(size):

        result.append([None] * size)

    for i in range(size):

        for j in range(i):

            (Pij, freqs) = csd(data[i], data[j], **params)
            result[i][j] = Pij
            result[j][i] = np.conj(Pij)

        (Pii, freqs) = psd(data[i], **params)
        result[i][i] = Pii

    return (np.array(result), freqs)


def sleeman( cov_mat, freq ):

    CHANNELS_COUNT = 3

    shape = cov_mat.shape

    if shape[0] != shape[1] or len(shape) != 3:

        raise ValueError("wrong matrix dimensions")

    if shape[0] != CHANNELS_COUNT:

        raise NotImplementedError("no generalization for this case. len(cov_mat) must be equal to 3!")

    omega_sq = square(2 * pi * freq)

    def get_noise( ch_id ):

        # Get two other channel id's
        second_id = (ch_id + 1) % 3
        third_id = (ch_id + 2) % 3

        return omega_sq*(cov_mat[ch_id][ch_id] - 
                         cov_mat[second_id][ch_id] * cov_mat[ch_id][third_id] /
                         cov_mat[second_id][third_id])

    return list(map(get_noise, range(CHANNELS_COUNT)))


def holcomb( freq, cov_mat, hs ):

    assert(cov_mat.shape[0] == len(hs) and cov_mat.shape[1] == len(hs))
    assert(cov_mat.shape[2] == len(freq))

    CHANNELS_COUNT = len(hs)

    def get_noise( ch_id ):

        other_id = (ch_id + 1) % CHANNELS_COUNT

        result = np.array(square(2 * pi * freq), dtype = np.complex)

        result *= (cov_mat[ch_id][ch_id] / (hs[ch_id] * conj(hs[ch_id])) -
                   cov_mat[other_id][ch_id] / (hs[ch_id] * conj(hs[other_id])))

        return result

    return list(map(get_noise, range(CHANNELS_COUNT)))
